from fdrtd.plugins.simon.caches.functional import CacheFunctional


class CacheAdditive(CacheFunctional):

    def __init__(self, minimum=1, maximum=-1):
        super().__init__(function=lambda x, y: x+y,
                         minimum=minimum,
                         maximum=maximum)



from fdrtd.plugins.simon.accumulators.accumulator import Accumulator


class AccumulatorGenericDictionary(Accumulator):

    def __init__(self, _=None):
        self.dictionary = {}

    def serialize(self):
        return {'dictionary': self.dictionary}

    @staticmethod
    def deserialize(dictionary):
        accumulator = AccumulatorGenericDictionary()
        accumulator.dictionary = dictionary['dictionary']
        return accumulator

    def add(self, other):
        self.dictionary = {**self.dictionary, **other}

    def update(self, data):
        key, value = data
        self.dictionary[key] = value

    def finalize(self):
        pass

    def get_dictionary(self):
        return self.dictionary



import secrets as _secrets

from fdrtd.plugins.simon.caches.cache import Cache
from fdrtd.plugins.simon.caches.additive import CacheAdditive
from fdrtd.plugins.simon.caches.functional import CacheFunctional
from fdrtd.plugins.simon.microprotocols.microprotocol import Microprotocol
from fdrtd.plugins.simon.accumulators.accumulator_statistics_frequency import AccumulatorStatisticsFrequency


class MicroprotocolStatisticsFrequency(Microprotocol):

    def __init__(self, microservice, properties, myself):
        super().__init__(microservice, properties, myself)

        self.secret_key = _secrets.token_bytes(32)

        self.n = self.network.count

        self.register_cache('samples', CacheAdditive(minimum=self.n))
        self.register_cache('input', Cache())
        self.register_cache('keys', CacheFunctional(lambda x, y: set(x).union(set(y)), self.n, self.n))
        self.register_cache('checkpoint2', CacheAdditive(minimum=self.n))
        self.register_cache('checkpoint3', CacheAdditive(minimum=self.n))

        self.register_stage(0, ['input'], self.stage_0)
        self.register_stage(1, ['keys'], self.stage_1)
        self.register_stage(2, ['checkpoint2'], self.stage_2)
        self.register_stage(3, ['checkpoint3'], self.stage_3)

        self.input = None
        self.result_cache = AccumulatorStatisticsFrequency()

        self.histogram = {}
        self.keys = []
        self.tags = {}

    def stage_0(self, args):

        self.network.broadcast(args['input']['samples'], 'samples')

        self.histogram = args['input']['histogram']
        self.network.broadcast([*self.histogram.keys()], 'keys')
        return 1, None

    def stage_1(self, args):
        self.keys = args['keys']
        for key in args['keys']:
            s = self.histogram.get(key, 0)
            tag = 'sum_{}'.format(key)
            self.tags[tag] = s
            self.register_cache(tag, Cache())
        self.network.broadcast(0, 'checkpoint2')
        return 2, None

    def stage_2(self, args):
        self.register_stage(4, ['samples', *self.tags.keys()], self.stage_final)
        self.network.broadcast(0, 'checkpoint3')
        return 3, None

    def stage_3(self, args):
        for tag in self.tags:
            self.start_pipeline('SecureSum', tag, self.tags[tag])
        return 4, None

    def stage_final(self, args):

        histogram = {}
        mode = None
        max = -1
        for tag in args:
            if tag == 'stage':
                continue
            if tag == 'samples':
                continue
            histogram[tag[4:]] = args[tag]['sum']
            if args[tag]['sum'] > max:
                max = args[tag]['sum']
                mode = tag[4:]

        return -1, {'inputs': self.n, 'result': {
                       'samples': args['samples'],
                       'mode': mode,
                       'histogram': histogram}}



from fdrtd.plugins.simon.accumulators.accumulator_basic_function import AccumulatorBasicFunction


class AccumulatorSecureSum(AccumulatorBasicFunction):

    def __init__(self, _=None):
        super().__init__(0, lambda x, y: x+y)

    def serialize(self):
        return {'samples': self.samples,
                'sum': self.data}

    @staticmethod
    def deserialize(dictionary):
        accumulator = AccumulatorSecureSum()
        accumulator.samples = dictionary['samples']
        accumulator.data = dictionary['sum']
        return accumulator

    def get_sum(self):
        return self.data



from representation import Api


class SyncApi(Api):

    def join_barrier(self, parties, party, uuid):
        ms = self.create(plugin="Sync", microservice='Barrier')
        sync = ms.create(uuid=uuid)
        sync.arrive(party=party)
        while sync.arrived() < parties:
            pass
        sync.depart(party=party)
        while sync.departed() < parties:
            pass
        return sync.reset()

    def send_broadcast(self, message, uuid):
        ms = self.create(plugin="Sync", microservice='Broadcast')
        return ms.send(uuid=uuid, message=message)

    def receive_broadcast(self, uuid):
        ms = self.create(plugin="Sync", microservice='Broadcast')
        rec = ms.receive(uuid=uuid)
        if rec is None:
            return None
        return self.download(rec)

    def clear_broadcast(self, uuid):
        ms = self.create(plugin="Sync", microservice='Broadcast')
        return ms.delete(uuid=uuid)

    def wait_for_broadcast(self, uuid):
        response = None
        while not response:
            response = self.receive_broadcast(uuid=uuid)
        return response



import fdrtd
import fdrtd.server.exceptions


def create(microprotocol):

    if microprotocol == 'MinimumMaximum':
        from fdrtd.plugins.simon.microprotocols.microprotocol_minimum_maximum import MicroprotocolMinimumMaximum
        return MicroprotocolMinimumMaximum

    if microprotocol == 'SecureSum':
        from fdrtd.plugins.simon.microprotocols.microprotocol_secure_sum import MicroprotocolSecureSum
        return MicroprotocolSecureSum

    if microprotocol == 'SecureMatrixMultiplication':
        from fdrtd.plugins.simon.microprotocols.microprotocol_secure_matrix_multiplication import MicroprotocolSecureMatrixMultiplication
        return MicroprotocolSecureMatrixMultiplication

    if microprotocol == 'SetIntersection':
        from fdrtd.plugins.simon.microprotocols.microprotocol_set_intersection import MicroprotocolSetIntersection
        return MicroprotocolSetIntersection

    if microprotocol == 'SetIntersectionSize':
        from fdrtd.plugins.simon.microprotocols.microprotocol_set_intersection_size import MicroprotocolSetIntersectionSize
        return MicroprotocolSetIntersectionSize

    if microprotocol == 'StatisticsBivariate':
        from fdrtd.plugins.simon.microprotocols.microprotocol_statistics_bivariate import MicroprotocolStatisticsBivariate
        return MicroprotocolStatisticsBivariate

    if microprotocol == 'StatisticsFrequency':
        from fdrtd.plugins.simon.microprotocols.microprotocol_statistics_frequency import MicroprotocolStatisticsFrequency
        return MicroprotocolStatisticsFrequency

    if microprotocol == 'StatisticsContingency':
        from fdrtd.plugins.simon.microprotocols.microprotocol_statistics_contingency import MicroprotocolStatisticsContingency
        return MicroprotocolStatisticsContingency

    if microprotocol == 'StatisticsUnivariate':
        from fdrtd.plugins.simon.microprotocols.microprotocol_statistics_univariate import MicroprotocolStatisticsUnivariate
        return MicroprotocolStatisticsUnivariate

    if microprotocol == 'StatisticsContingencyVertical':
        from fdrtd.plugins.simon.microprotocols.microprotocol_statistics_contingency_vertical import MicroprotocolStatisticsContingencyVertical
        return MicroprotocolStatisticsContingencyVertical

    if microprotocol == 'StatisticsRegressionOLSVertical':
        from fdrtd.plugins.simon.microprotocols.microprotocol_statistics_regression_ols_vertical import MicroprotocolStatisticsRegressionOLSVertical
        return MicroprotocolStatisticsRegressionOLSVertical

    raise fdrtd.server.exceptions.NotAvailable(microprotocol)



import numpy as _numpy

from fdrtd.plugins.simon.caches.cache import Cache
from fdrtd.plugins.simon.microprotocols.microprotocol import Microprotocol


class MicroprotocolStatisticsRegressionOLSVertical(Microprotocol):

    def __init__(self, microservice, properties, myself):
        super().__init__(microservice, properties, myself)

        self.register_cache('input', Cache())
        self.register_cache('final', Cache())

        self.register_stage(0, ['input'], self.stage_0)
        self.register_stage(1, ['final'], self.stage_1)

    def stage_0(self, args):
        if self.network.myself == 0:
            x = _numpy.array(args['input'])
            xt = x.transpose()
            n = _numpy.matmul(xt, x)
            ninv = _numpy.linalg.inv(n)
            mp = _numpy.matmul(ninv, xt)
            self.start_pipeline('SecureMatrixMultiplication', 'final', [mp.tolist()])
        else:
            self.start_pipeline('SecureMatrixMultiplication', 'final', [[[x] for x in args['input']]])
        return 1, None

    def stage_1(self, args):
        return -1, {'inputs': 2,  # self.n,
                    'result': {
                        'mle': args['final']['product']}}



import numpy as _numpy

from fdrtd.plugins.simon.caches.cache import Cache
from fdrtd.plugins.simon.microprotocols.microprotocol import Microprotocol


class MicroprotocolSecureMatrixMultiplication(Microprotocol):

    def __init__(self, microservice, properties, myself):
        super().__init__(microservice, properties, myself)

        self.register_cache('input', Cache())
        self.register_cache('dimX', Cache())
        self.register_cache('dimY', Cache())
        self.register_cache('intermediateV', Cache())
        self.register_cache('intermediateW', Cache())
        self.register_cache('final', Cache())

        self.register_stage(0, ['input'], self.stage_0)
        self.register_stage(1, ['dimX', 'dimY'], self.stage_1)
        self.register_stage(2, ['intermediateV'], self.stage_2)
        self.register_stage(3, ['intermediateW'], self.stage_3)
        self.register_stage(4, ['final'], self.stage_4)

        self.M = None
        self.n = self.p = self.q = 0
        self.v = None

    def stage_0(self, args):
        self.M = _numpy.array(args['input'])
        self.network.broadcast(self.M.shape, 'dimX' if self.network.myself == 0 else 'dimY')
        return 1, None

    def stage_1(self, args):
        self.p = args['dimX'][0]
        self.n = args['dimX'][1]
        if args['dimY'][0] != args['dimX'][1]:
            raise RuntimeError("matrix shapes not compatible")
        self.q = args['dimY'][1]

        if self.network.myself == 0:
            xt = self.M.transpose()
            q, r = _numpy.linalg.qr(xt, mode='complete')
            g = int((self.p*self.n)/(self.p+self.q))
            z = q[:, self.p:self.p+g]
            self.v = _numpy.identity(self.n) - _numpy.matmul(z, z.transpose())
            self.network.broadcast(self.v.tolist(), 'intermediateV')

        return 2, None

    def stage_2(self, args):

        if self.network.myself == 1:
            v = args['intermediateV']
            w = _numpy.matmul(v, self.M)
            self.network.broadcast(w.tolist(), 'intermediateW')

        return 3, None

    def stage_3(self, args):

        if self.network.myself == 0:
            w = args['intermediateW']
            result = _numpy.matmul(self.M, w)
            self.network.broadcast(result.tolist(), 'final')

        return 4, None

    def stage_4(self, args):
        return -1, {'inputs': 2,  # self.n,
                    'result': {
                        'product': args['final']}}



import uuid as _uuid

from fdrtd.plugins.simon.peer_to_peer import PeerToPeer
import representation


class Microprotocol:

    def __init__(self, microservice, properties, myself):

        self.microservice = microservice
        self.uuid = properties['task_id']
        self.network = PeerToPeer({'namespace': "fdrtd", 'protocol': "Simon"},
                                  properties['task_id'], properties, myself)
        self.stage = 0
        self.stages = {}
        self.caches = {}
        self.result = None
        self.parameters = properties['parameters']
        if properties['parent'] is None:
            self.parent_id = None
            self.parent_token = None
        else:
            self.parent_id = properties['parent']['parent_id']
            self.parent_token = properties['parent']['parent_token']
        self.myself = myself

    def register_cache(self, argument, cache):
        self.caches[argument] = cache

    def register_stage(self, stage, required_arguments, handler):
        self.stages[stage] = (required_arguments, handler)

    def process(self, message):
        self.caches[message['token']].process(message['body'])
        self.continue_processing()

    def continue_processing(self):

        if self.stage not in self.stages:
            return

        (required_arguments, handler) = self.stages[self.stage]

        arguments = {'stage': self.stage}
        for argument in required_arguments:
            if argument not in self.caches:
                break
            arg = self.caches[argument].get_data()
            if arg is None:
                break
            arguments[argument] = arg
        else:
            self.stage = -1
            stage_after, ret = handler(arguments)
            if stage_after == -1:
                self.result = ret
                if self.parent_id is not None:
                    message = {'token': self.parent_token, 'sender': self.network.myself, 'receiver': self.network.myself, 'body': ret['result']}
                    task = self.microservice.get_task(task_id=self.parent_id)
                    task.microprotocol.process(message)
            else:
                self.stage = stage_after
                self.continue_processing()

    def get_result(self):
        return self.result

    def start_pipeline(self, microprotocol, token, body):
        network = {'nodes': self.network.nodes, 'myself': self.network.myself}
        task = self.derive_task(network, microprotocol, self.uuid, token)
        task.input(data=body)
        task.start()

    def derive_task(self, network, microprotocol, parent_id, token):
        node = network['nodes'][network['myself']]
        if isinstance(node, str):
            api = representation.Api(url=node)
        else:
            api = representation.Api(interface=node)  # for local testing
        microservice = api.create(namespace="fdrtd", protocol="Simon")
        return microservice.create_task(microprotocol=microprotocol, network=network,
                                        parent={'parent_id': parent_id, 'parent_token': token},
                                        task_id=str(_uuid.uuid5(_uuid.UUID(parent_id), token)))



from fdrtd.plugins.simon.accumulators.accumulator import Accumulator


class AccumulatorBasicFunction(Accumulator):

    def __init__(self, initial, function):
        self.samples = 0
        self.data = initial
        self.function = function

    def add(self, other):
        self.samples = self.samples + other.samples
        self.data = self.function(self.data, other.data)

    def update(self, data):
        if isinstance(data, list):
            self.samples = self.samples + len(data)
            for datum in data:
                self.data = self.function(self.data, datum)
        else:
            self.samples = self.samples + 1
            self.data = self.function(self.data, data)

    def finalize(self):
        pass

    def get_samples(self):
        return self.samples

    def get_data(self):
        return self.data



from fdrtd.plugins.simon.caches.cache import Cache


class CacheFunctional(Cache):

    def __init__(self, function, minimum=1, maximum=-1):
        super().__init__()
        self.function = function
        self.intermediate = None
        self.count = 0
        self.minimum = minimum
        self.maximum = maximum

    def process(self, data):
        if self.count < self.maximum or self.maximum == -1:
            if self.intermediate is None:
                self.intermediate = data
            else:
                self.intermediate = self.function(self.intermediate, data)
            self.count = self.count + 1
        if self.count >= self.minimum:
            self.data = self.intermediate



import representation


class PeerToPeer:

    def __init__(self, endpoint, task_id, properties, myself):
        self.endpoint = endpoint
        self.task_id = task_id
        self.properties = properties
        self.nodes = properties['nodes']
        self.myself = myself
        self.count = len(self.nodes)

    def broadcast(self, body, token):
        for index in range(self.count):
            self.send_to_node(body, index, token)

    def send_to_next_node(self, body, token):
        index = (self.myself + 1) % len(self.nodes)
        self.send_to_node(body, index, token)

    def send_to_previous_node(self, body, token):
        index = (self.myself - 1) % len(self.nodes)
        self.send_to_node(body, index, token)

    def send_to_node(self, body, receiver, token):
        if isinstance(self.nodes[receiver], str):
            api = representation.Api(url=self.nodes[receiver])
        else:
            api = representation.Api(interface=self.nodes[receiver])  # for testing purposes
        microservice = api.create(**self.endpoint)
        microservice.peer_to_peer(callback=self.task_id,
                                  body={
                                      'task': self.properties,
                                      'message': {
                                          'token': token,
                                          'sender': self.myself,
                                          'receiver': receiver,
                                          'body': body
                                      }
                                  })



from fdrtd.plugins.simon.caches.cache import Cache
from fdrtd.plugins.simon.microprotocols.microprotocol import Microprotocol
from fdrtd.plugins.simon.accumulators.accumulator_statistics_bivariate import AccumulatorStatisticsBivariate


class MicroprotocolStatisticsBivariate(Microprotocol):

    def __init__(self, microservice, properties, myself):
        super().__init__(microservice, properties, myself)

        self.n = self.network.count

        self.digits_before = properties['parameters'].get('digits_before', 12)
        self.digits_after = properties['parameters'].get('digits_after', 12)

        self.register_cache('input', Cache())
        self.register_cache('x_uncentered_0', Cache())
        self.register_cache('x_uncentered_1', Cache())
        self.register_cache('x_uncentered_2', Cache())
        self.register_cache('x_uncentered_3', Cache())
        self.register_cache('x_uncentered_4', Cache())
        self.register_cache('x_uncentered_5', Cache())
        self.register_cache('x_samples', Cache())
        self.register_cache('x_geometric', Cache())
        self.register_cache('x_harmonic', Cache())
        self.register_cache('x_minmax', Cache())
        self.register_cache('y_uncentered_0', Cache())
        self.register_cache('y_uncentered_1', Cache())
        self.register_cache('y_uncentered_2', Cache())
        self.register_cache('y_uncentered_3', Cache())
        self.register_cache('y_uncentered_4', Cache())
        self.register_cache('y_uncentered_5', Cache())
        self.register_cache('y_samples', Cache())
        self.register_cache('y_geometric', Cache())
        self.register_cache('y_harmonic', Cache())
        self.register_cache('y_minmax', Cache())
        self.register_cache('xy_uncentered_0', Cache())
        self.register_cache('xy_uncentered_1', Cache())
        self.register_cache('xy_uncentered_2', Cache())
        self.register_cache('xy_uncentered_3', Cache())
        self.register_cache('xy_uncentered_4', Cache())
        self.register_cache('xy_uncentered_5', Cache())
        self.register_cache('xy_samples', Cache())
        self.register_cache('xy_geometric', Cache())
        self.register_cache('xy_harmonic', Cache())
        self.register_cache('xy_minmax', Cache())

        self.register_stage(0, ['input'], self.stage_0)
        self.register_stage(1, ['x_uncentered_0', 'x_uncentered_1', 'x_uncentered_2', 'x_uncentered_3', 'x_uncentered_4',
                                'x_uncentered_5', 'x_samples', 'x_geometric', 'x_harmonic', 'x_minmax'], self.stage_x)
        self.register_stage(2, ['y_uncentered_0', 'y_uncentered_1', 'y_uncentered_2', 'y_uncentered_3', 'y_uncentered_4',
                                'y_uncentered_5', 'y_samples', 'y_geometric', 'y_harmonic', 'y_minmax'], self.stage_y)
        self.register_stage(3, ['xy_uncentered_0', 'xy_uncentered_1', 'xy_uncentered_2', 'xy_uncentered_3', 'xy_uncentered_4',
                                'xy_uncentered_5', 'xy_samples', 'xy_geometric', 'xy_harmonic', 'xy_minmax'], self.stage_xy)

        self.input = None
        self.result_cache = AccumulatorStatisticsBivariate()

    def start_pipelines(self, prefix, accumulator):
        self.start_pipeline('SecureSum', prefix+'uncentered_0', accumulator.moments.uncentered[0])
        self.start_pipeline('SecureSum', prefix+'uncentered_1', accumulator.moments.uncentered[1])
        self.start_pipeline('SecureSum', prefix+'uncentered_2', accumulator.moments.uncentered[2])
        self.start_pipeline('SecureSum', prefix+'uncentered_3', accumulator.moments.uncentered[3])
        self.start_pipeline('SecureSum', prefix+'uncentered_4', accumulator.moments.uncentered[4])
        self.start_pipeline('SecureSum', prefix+'uncentered_5', accumulator.moments.uncentered[5])
        self.start_pipeline('SecureSum', prefix+'samples', accumulator.samples)
        self.start_pipeline('SecureSum', prefix+'geometric', accumulator.geometric)
        self.start_pipeline('SecureSum', prefix+'harmonic', accumulator.harmonic)
        self.start_pipeline('MinimumMaximum', prefix+'minmax', [{'minimum': accumulator.minimum, 'maximum': accumulator.maximum}])

    def stage_0(self, args):
        self.input = AccumulatorStatisticsBivariate.deserialize(args['input'])
        self.start_pipelines('x_', self.input.accumulator_x)
        self.start_pipelines('y_', self.input.accumulator_y)
        self.start_pipelines('xy_', self.input.accumulator_xy)
        return 1, None

    def stage_z(self, args, prefix, accumulator):
        accumulator.moments.uncentered[0] = args[prefix+'uncentered_0']['sum']
        accumulator.moments.uncentered[1] = args[prefix+'uncentered_1']['sum']
        accumulator.moments.uncentered[2] = args[prefix+'uncentered_2']['sum']
        accumulator.moments.uncentered[3] = args[prefix+'uncentered_3']['sum']
        accumulator.moments.uncentered[4] = args[prefix+'uncentered_4']['sum']
        accumulator.moments.uncentered[5] = args[prefix+'uncentered_5']['sum']
        accumulator.samples = args[prefix+'samples']['sum']
        accumulator.geometric = args[prefix+'geometric']['sum']
        accumulator.harmonic = args[prefix+'harmonic']['sum']
        accumulator.minimum = args[prefix+'minmax']['minimum']
        accumulator.maximum = args[prefix+'minmax']['maximum']
        accumulator.moments.samples = accumulator.samples
        accumulator.moments.center = 0.0
        accumulator.moments.centered = [x for x in accumulator.moments.uncentered]
        accumulator.moments.set_center(accumulator.moments.uncentered[0] / accumulator.samples)

    def stage_x(self, args):
        self.stage_z(args, "x_", self.result_cache.accumulator_x)
        return 2, None

    def stage_y(self, args):
        self.stage_z(args, "y_", self.result_cache.accumulator_y)
        return 3, None

    def stage_xy(self, args):
        self.stage_z(args, "xy_", self.result_cache.accumulator_xy)
        return -1, {'inputs': self.n, 'result': self.result_cache.evaluate_to_dict()}



import secrets as _secrets

from fdrtd.plugins.simon.caches.cache import Cache
from fdrtd.plugins.simon.caches.additive import CacheAdditive
from fdrtd.plugins.simon.microprotocols.microprotocol import Microprotocol


class MicroprotocolSecureSum(Microprotocol):

    def __init__(self, microservice, properties, myself):
        super().__init__(microservice, properties, myself)

        self.n = self.network.count

        self.digits_before = properties['parameters'].get('digits_before', 12)
        self.digits_after = properties['parameters'].get('digits_after', 12)

        self.intermediate = 0

        self.register_cache('input', Cache())
        self.register_cache('samples', CacheAdditive(minimum=self.n))
        self.register_cache('intermediate', CacheAdditive(minimum=self.n))
        self.register_cache('final', CacheAdditive(minimum=self.n))

        self.register_stage(0, ['input'], self.stage_0)
        self.register_stage(1, ['intermediate'], self.stage_1)
        self.register_stage(2, ['final'], self.stage_2)

    def stage_0(self, args):
        self.network.broadcast(args['input']['samples'], 'samples')
        o = args['input']['sum']
        s = 0
        for i in range(self.network.count - 1):
            r = _secrets.randbelow(10 ** (self.digits_after+self.digits_before))
            s = s + r
            self.network.send_to_node(r, i, 'intermediate')
        self.network.send_to_node(int(o * (10 ** self.digits_after)) - s, self.network.count - 1, 'intermediate')
        return 1, None

    def stage_1(self, args):
        self.network.broadcast(args['intermediate'], 'final')
        return 2, None

    def stage_2(self, args):
        return -1, {'inputs': self.n,
                    'result': {
                        'samples': self.caches['samples'].get_data(),
                        'sum': args['final'] / (10 ** self.digits_after)}}



import math as _math
import numpy as _numpy

from fdrtd.plugins.simon.caches.cache import Cache
from fdrtd.plugins.simon.microprotocols.microprotocol import Microprotocol


class MicroprotocolStatisticsRegressionLogitVertical(Microprotocol):

    def __init__(self, microservice, properties, myself):
        super().__init__(microservice, properties, myself)

        self.register_cache('input', Cache())
        self.register_cache('final', Cache())

        self.register_stage(0, ['input'], self.stage_0)
        self.register_stage(1, ['final'], self.stage_1)

    def stage_0(self, args):
        if self.network.myself == 0:
            n = len(args['input'])
            m = len(args['input'][0])
            regressor = [[1, *row] for row in args['input']]
            x = _numpy.array(regressor)
            xt = x.transpose()
            w0 = [0.0 for i in range(m+1)]
            mu = [1.0 / _math.exp(-sum(w0[k] * x[i][k] for k in range(m+1))) for i in range(n)]
            s = _numpy.diag(mu)

            xts = _numpy.matmul(xt, s)
            xtsx = _numpy.matmul(xts, x)
            xtsxi = _numpy.linalg.inv(xtsx)
            xtsxixt = _numpy.matmul(xtsxi, xt)

            sx = _numpy.matmul(s, x)
            sxw = _numpy.matmul(sx, w0)

            sxwymu = sxw + _numpy.array([0, 0, 0, 0, 0, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 1, 1, 1, 1, 1]) - mu

            self.w1 = _numpy.matmul(xtsxixt, sxwymu)

            #n = _numpy.matmul(xt, x)
            #ninv = _numpy.linalg.inv(n)
            #mp = _numpy.matmul(ninv, xt)
            #self.start_pipeline('SecureMatrixMultiplication', 'final', [mp.tolist()])
        else:
            pass
            #self.start_pipeline('SecureMatrixMultiplication', 'final', [[[x] for x in args['input']]])
        return 1, None

    def stage_1(self, args):
        return -1, {'inputs': 2,  # self.n,
                    'result': {
                        'mle': self.w1
                    #    args['final']['product']
                    }}



from fdrtd.plugins.simon.accumulators.accumulator import Accumulator


class AccumulatorGeneric(Accumulator):

    def __init__(self, _=None):
        self.data = None

    def serialize(self):
        return self.data

    @staticmethod
    def deserialize(data):
        accumulator = AccumulatorGeneric()
        accumulator.data = data
        return accumulator

    def add(self, other):
        self.data = other.data
        return self

    def update(self, data):
        self.data = data

    def finalize(self):
        pass

    def get_data(self):
        return self.data



# the following code is in part derived from 'Code Components' published in:
# A. Langley, M. Hamburg, S. Turner: Elliptic Curves for Security
# Internet Research Task Force (IRTF) Request for Comments 7748 (January 2016)
# for the original document and code components see https://tools.ietf.org/html/rfc7748
# for licensing information see https://trustee.ietf.org/trust-legal-provisions.html:
# [quote] Code Components are also licensed to each person who wishes to receive such
# a license on the terms of the “Simplified BSD License", as described below [...]
#
# BSD License:
# Copyright (c) <insert year> IETF Trust and the persons identified as authors of the code. All rights reserved.
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
# * Redistributions of source code must retain the above copyright notice,
#   this list of conditions and the following disclaimer.
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
# * Neither the name of Internet Society, IETF or IETF Trust, nor the names of specific contributors,
#   may be used to endorse or promote products derived from this software without specific prior written permission.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS “AS IS” AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
# USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


class X25519X448:

    def __init__(self, bits, p, a24, masked_key):
        self.bits = bits
        self.p = p
        self.a24 = a24
        self.mask = (1 << self.bits) - 1
        self.decoded_key = self.decode_little_endian(masked_key)

    def decode_little_endian(self, b):
        return sum([b[i] << 8 * i for i in range((self.bits + 7) // 8)])

    def decode_u_coordinate(self, u):
        u_list = [b for b in u]
        if self.bits % 8:
            u_list[-1] &= (1 << (self.bits % 8)) - 1
        return self.decode_little_endian(u_list)

    def encode_u_coordinate(self, u):
        u = u % self.p
        return bytes([((u >> 8 * i) & 0xff) for i in range((self.bits + 7) // 8)])

    def montgomery(self, k, u):
        x_1 = u
        x_2 = 1
        z_2 = 0
        x_3 = u
        z_3 = 1
        swap = 0
        for t in range(self.bits-1, -1, -1):
            k_t = (k >> t) & 1
            swap ^= k_t
            (x_2, x_3) = self.cswap(swap, x_2, x_3)
            (z_2, z_3) = self.cswap(swap, z_2, z_3)
            swap = k_t
            a = (x_2 + z_2) % self.p
            aa = pow(a, 2, self.p)
            b = (x_2 - z_2) % self.p
            bb = pow(b, 2, self.p)
            e = (aa - bb) % self.p
            c = (x_3 + z_3) % self.p
            d = (x_3 - z_3) % self.p
            da = (d * a) % self.p
            cb = (c * b) % self.p
            x_3 = pow(da + cb, 2, self.p)
            z_3 = (x_1 * pow(da - cb, 2, self.p)) % self.p
            x_2 = (aa * bb) % self.p
            z_2 = (e * (aa + self.a24 * e)) % self.p
        (x_2, x_3) = self.cswap(swap, x_2, x_3)
        (z_2, z_3) = self.cswap(swap, z_2, z_3)
        return (x_2 * pow(z_2, self.p - 2, self.p)) % self.p

    def cswap(self, swap, x_2, x_3):
        if swap:
            mask = self.mask
        else:
            mask = 0
        dummy = mask & (x_2 ^ x_3)
        x_2 = x_2 ^ dummy
        x_3 = x_3 ^ dummy
        return x_2, x_3

    def encrypt(self, data):
        k = self.decoded_key
        u = self.decode_u_coordinate(data)
        t = self.montgomery(k, u)
        return self.encode_u_coordinate(t)


class X25519(X25519X448):

    def __init__(self, secret_key):
        super().__init__(masked_key=self.mask_key(secret_key),
                         bits=255, p=2 ** 255 - 19, a24=121665)

    @staticmethod
    def mask_key(secret_key):
        masked_key = [b for b in secret_key]
        masked_key[0] &= 248
        masked_key[31] &= 127
        masked_key[31] |= 64
        return masked_key


class X448(X25519X448):

    def __init__(self, secret_key):
        super().__init__(masked_key=self.mask_key(secret_key),
                         bits=448, p=2 ** 448 - 2 ^ 224 - 1, a24=39081)

    @staticmethod
    def mask_key(secret_key):
        masked_key = [b for b in secret_key]
        masked_key[0] &= 252
        masked_key[55] |= 128
        return masked_key



from fdrtd.plugins.simon.accumulators.accumulator import Accumulator


class AccumulatorStatisticsFrequency(Accumulator):

    def __init__(self, _=None):
        self.samples = 0
        self.histogram = {}
        self.mode = None

    def serialize(self):
        return {'samples': self.samples,
                'mode': self.mode,
                'histogram': self.histogram}

    @staticmethod
    def deserialize(dictionary):
        accumulator = AccumulatorStatisticsFrequency()
        accumulator.samples = dictionary['samples']
        accumulator.mode = dictionary['mode']
        accumulator.histogram = dictionary['histogram']
        return accumulator

    def add(self, other):
        self.samples += other.samples
        for item in other.histogram:
            self.histogram[item] = self.histogram.get(item, 0) + other.histogram[item]

    def update(self, key, value=1):
        self.samples = self.samples + value
        self.histogram[key] = self.histogram.get(key, 0) + value

    def finalize(self):
        maximum = 0
        for item in self.histogram:
            if self.histogram[item] > maximum:
                maximum = self.histogram[item]
                self.mode = item

    def get_histogram(self):
        return self.histogram

    def get_samples(self):
        return self.samples

    def get_mode(self):
        return self.mode



import math as _math
import sys as _sys

from fdrtd.plugins.simon.accumulators.accumulator import Accumulator
from fdrtd.plugins.simon.accumulators.accumulator_statistics_moments import AccumulatorStatisticsMoments


class AccumulatorStatisticsUnivariate(Accumulator):

    def __init__(self, _=None):
        self.moments = AccumulatorStatisticsMoments()
        self.samples = 0
        self.minimum = _sys.float_info.max
        self.maximum = _sys.float_info.min
        self.geometric = 0.0
        self.harmonic = 0.0

    def serialize(self):
        return {'moments': self.moments.serialize(),
                'samples': self.samples,
                'minimum': self.minimum,
                'maximum': self.maximum,
                'geometric': self.geometric,
                'harmonic': self.harmonic}

    @staticmethod
    def deserialize(dictionary):
        accumulator = AccumulatorStatisticsUnivariate()
        accumulator.moments = AccumulatorStatisticsMoments.deserialize(dictionary['moments'])
        accumulator.samples = dictionary['samples']
        accumulator.minimum = dictionary['minimum']
        accumulator.maximum = dictionary['maximum']
        accumulator.geometric = dictionary['geometric']
        accumulator.harmonic = dictionary['harmonic']
        return accumulator

    def add(self, other):
        self.moments.add(other.moments)
        self.samples += other.samples
        if self.minimum > other.minimum:
            self.minimum = other.minimum
        if self.maximum < other.maximum:
            self.maximum = other.maximum
        self.geometric += other.geometric
        self.harmonic += other.harmonic

    def update(self, data):
        self.moments.update(data)
        self.samples += 1
        if self.minimum > data:
            self.minimum = data
        if self.maximum < data:
            self.maximum = data
        if data > 0.0:
            self.geometric += _math.log(data)
        if data != 0.0:
            self.harmonic += 1.0 / data

    def finalize(self):
        self.moments.set_center(self.get_mean())

    def encrypt_data_for_upload(self, nonce, power=1):
        return {'moments': self.moments.encrypt_data_for_upload(nonce, power=power),
                'samples': self.samples,
                'minimum': nonce.encrypt_numerical(self.minimum, power=power),
                'maximum': nonce.encrypt_numerical(self.maximum, power=power),
                'geometric': self.geometric + _math.log(nonce.numerical) * power * self.samples,
                'harmonic': self.harmonic / pow(nonce.numerical, power)}

    @staticmethod
    def decrypt_result_from_download(encrypted, nonce):
        decryption_powers = {'minimum': 1, 'maximum': 1, 'sum': 1,
                             'mean': 1, 'harmonic_mean': 1, 'geometric_mean': 1,
                             'variance': 2, 'variance_mle': 2, 'variance_of_sample_mean': 2,
                             'standard_deviation': 1, 'standard_deviation_mle': 1, 'standard_error_of_sample_mean': 1,
                             'coefficient_of_variation': 0, 'coefficient_of_variation_mle': 0,
                             'root_mean_square': 1, 'root_mean_square_deviation': 1, 'skewness': 0,
                             'kurtosis': 0, 'kurtosis_excess': 0, 'hyper_skewness': 0, 'hyper_flatness': 0}
        return nonce.decrypt_dictionary_numerical(encrypted, decryption_powers)

    def calculate_raw_moment(self, k):
        if k == 0:
            return 1.0
        return self.moments.uncentered[k - 1] / self.samples

    def calculate_central_moment(self, k):
        if k == 0:
            return 1.0
        if k == 1:
            return 0.0
        return self.moments.centered[k - 1] / self.samples

    def calculate_standardized_moment(self, k):
        if k == 0:
            return 1.0
        return self.calculate_central_moment(k) * pow(self.get_variance_mle(), -0.5 * k)

    def calculate_raw_moments(self):
        return [self.calculate_raw_moment(k) for k in range(6)]

    def calculate_central_moments(self):
        return [self.calculate_central_moment(k) for k in range(6)]

    def calculate_standardized_moments(self):
        return [self.calculate_standardized_moment(k) for k in range(6)]

    def calculate_center(self):
        return self.moments.center
    
    @staticmethod
    def list_outputs():
        return ['samples', 'minimum', 'maximum', 'sum', 'mean', 'harmonic_mean', 'geometric_mean',
                'variance', 'variance_mle', 'variance_of_sample_mean', 'standard_deviation', 'standard_deviation_mle',
                'standard_error_of_sample_mean', 'coefficient_of_variation', 'coefficient_of_variation_mle',
                'root_mean_square', 'root_mean_square_deviation',
                'skewness', 'kurtosis', 'kurtosis_excess', 'hyper_skewness', 'hyper_flatness']

    def get_samples(self):
        return self.samples

    def get_minimum(self):
        return self.minimum

    def get_maximum(self):
        return self.maximum

    def get_sum(self):
        return self.moments.uncentered[0]

    def get_mean(self):
        return self.moments.uncentered[0] / self.samples

    def get_harmonic_mean(self):
        return self.samples / self.harmonic

    def get_geometric_mean(self):
        return _math.exp(self.geometric / self.samples)

    def get_variance(self):
        return self.moments.centered[1] / (self.samples - 1.0)

    def get_variance_mle(self):
        return self.moments.centered[1] / self.samples

    def get_variance_of_sample_mean(self):
        return self.get_variance() / self.samples

    def get_standard_deviation(self):
        return _math.sqrt(self.get_variance())

    def get_standard_deviation_mle(self):
        return _math.sqrt(self.get_variance_mle())

    def get_standard_error_of_sample_mean(self):
        return _math.sqrt(self.get_variance_of_sample_mean())

    def get_coefficient_of_variation(self):
        return self.get_standard_deviation() / self.get_mean()

    def get_coefficient_of_variation_mle(self):
        return self.get_standard_deviation_mle() / self.get_mean()

    def get_root_mean_square(self):
        return _math.sqrt(self.calculate_raw_moment(2))

    def get_root_mean_square_deviation(self):
        return _math.sqrt(self.calculate_central_moment(2))

    def get_skewness(self):
        return self.calculate_standardized_moment(3)

    def get_kurtosis(self):
        return self.calculate_standardized_moment(4)

    def get_kurtosis_excess(self):
        return self.get_kurtosis() - 3.0

    def get_hyper_skewness(self):
        return self.calculate_standardized_moment(5)

    def get_hyper_flatness(self):
        return self.calculate_standardized_moment(6)



import uuid as _uuid
import time as _time

from fdrtd.plugins.simon.task import TaskSimon
from fdrtd.plugins.simon.sync_api import SyncApi


class MicroserviceSimon:

    def __init__(self):
        self.tasks = {}
        self._the_cache = []

    def get_task(self, task_id):
        if task_id in self.tasks:
            return self.tasks.get(task_id)
        else:
            return None

    def create_task(self, microprotocol, network, parameters=None, task_id=None, parent=None):

        if task_id is None:
            task_id = str(_uuid.uuid4())

        if parameters is None:
            parameters = {}

        task = TaskSimon(self, network, microprotocol, parameters, task_id, parent)

        for item in self._the_cache:
            if item['task_id'] == task_id:
                task.peer_to_peer(item['body'])

        self.tasks[task_id] = task
        return task

    def join_task(self, invitation, network):

        task_id = invitation['task_id']

        task = TaskSimon(self, network, invitation['microprotocol'], invitation['parameters'], task_id, invitation['parent'])

        for item in self._the_cache:
            if item['task_id'] == task_id:
                task.peer_to_peer(item['body'])

        self.tasks[task_id] = task
        return task

    def peer_to_peer(self, callback, body):
        if callback in self.tasks:
            self.tasks[callback].peer_to_peer(body)
        else:
            self._the_cache.append({'task_id': callback, 'body': body})
        return None

    def compute(self, microprotocol, data, network):
        sync_api = SyncApi(network['nodes'][0])
        if network['myself'] == 0:
            task = self.create_task(microprotocol, network)
            invitation = task.invite()
            sync_api.send_broadcast(invitation, network['uuid'])
        else:
            invitation = sync_api.wait_for_broadcast(network['uuid'])
            task = self.join_task(invitation, network)
        task.input(data)
        task.start()
        result = task.result()
        while result is None:
            _time.sleep(0.02)
            result = task.result()
        return result



from fdrtd.plugins.simon.accumulators.accumulator import Accumulator


class AccumulatorSetIntersection(Accumulator):

    def __init__(self, _=None):
        self.samples = 0
        self.set = set()

    def serialize(self):
        return {'samples': self.samples,
                'set': list(self.set)}

    @staticmethod
    def deserialize(dictionary):
        accumulator = AccumulatorSetIntersection()
        accumulator.samples = dictionary['samples']
        accumulator.set = set(dictionary['set'])
        return accumulator

    def add(self, other):
        if self.samples == 0:
            self.set = other.set
        else:
            self.set = self.set.intersection(other.set)
        self.samples = self.samples + other.samples

    def update(self, data):
        if self.samples == 0:
            self.set = set(data)
        else:
            self.set = self.set.intersection(set(data))
        self.samples = self.samples + 1

    def finalize(self):
        pass

    def get_samples(self):
        return self.samples

    def get_size_intersection(self):
        return len(self.set)

    def get_intersection(self):
        return list(self.set)



from fdrtd.plugins.simon.accumulators.accumulator import Accumulator


class AccumulatorStatisticsContingency(Accumulator):

    def __init__(self, _=None):
        self.samples = 0
        self.table = {}
        self.mode = None

    def serialize(self):
        return {'samples': self.samples,
                'mode': self.mode,
                'table': self.table}

    @staticmethod
    def deserialize(dictionary):
        accumulator = AccumulatorStatisticsContingency()
        accumulator.samples = dictionary['samples']
        accumulator.mode = dictionary['mode']
        accumulator.table = dictionary['table']
        return accumulator

    def add(self, other):
        self.samples += other.samples
        for histogram in other.table:
            if histogram not in self.table:
                self.table[histogram] = {}
            for item in other.table[histogram]:
                self.table[histogram][item] = self.table[histogram].get(item, 0) \
                                              + other.table[histogram][item]

    def update(self, data, count=1):
        (histogram, item) = data
        if histogram not in self.table:
            self.table[histogram] = {}
        self.table[histogram][item] = self.table[histogram].get(item, 0) + count
        self.samples = self.samples + count

    def finalize(self):
        maximum = 0
        for histogram in self.table:
            for item in self.table[histogram]:
                if self.table[histogram][item] > maximum:
                    maximum = self.table[histogram][item]
                    self.mode = (histogram, item)

    def get_samples(self):
        return self.samples

    def get_table(self):
        return self.table

    def get_mode(self):
        return self.mode



import fdrtd


def create(accumulator_name):

    if accumulator_name == 'MinimumMaximum':
        from fdrtd.plugins.simon.accumulators.accumulator_basic_minimum_maximum import AccumulatorBasicMinimumMaximum
        return AccumulatorBasicMinimumMaximum

    if accumulator_name == 'SecureSum':
        from fdrtd.plugins.simon.accumulators.accumulator_basic_sum import AccumulatorSecureSum
        return AccumulatorSecureSum

    if accumulator_name == 'SecureMatrixMultiplication':
        from fdrtd.plugins.simon.accumulators.accumulator_generic import AccumulatorGeneric
        return AccumulatorGeneric

    if accumulator_name == 'SetIntersection':
        from fdrtd.plugins.simon.accumulators.accumulator_set_intersection import AccumulatorSetIntersection
        return AccumulatorSetIntersection

    if accumulator_name == 'SetIntersectionSize':
        from fdrtd.plugins.simon.accumulators.accumulator_set_intersection_size import AccumulatorSetIntersectionSize
        return AccumulatorSetIntersectionSize

    if accumulator_name == 'StatisticsBivariate':
        from fdrtd.plugins.simon.accumulators.accumulator_statistics_bivariate import AccumulatorStatisticsBivariate
        return AccumulatorStatisticsBivariate

    if accumulator_name == 'StatisticsFrequency':
        from fdrtd.plugins.simon.accumulators.accumulator_statistics_frequency import AccumulatorStatisticsFrequency
        return AccumulatorStatisticsFrequency

    if accumulator_name == 'StatisticsContingency':
        from fdrtd.plugins.simon.accumulators.accumulator_statistics_contingency import AccumulatorStatisticsContingency
        return AccumulatorStatisticsContingency

    if accumulator_name == 'StatisticsUnivariate':
        from fdrtd.plugins.simon.accumulators.accumulator_statistics_univariate import AccumulatorStatisticsUnivariate
        return AccumulatorStatisticsUnivariate

    if accumulator_name == 'StatisticsContingencyVertical':
        from fdrtd.plugins.simon.accumulators.accumulator_generic_dictionary import AccumulatorGenericDictionary
        return AccumulatorGenericDictionary

    if accumulator_name == 'StatisticsRegressionOLSVertical':
        from fdrtd.plugins.simon.accumulators.accumulator_generic import AccumulatorGeneric
        return AccumulatorGeneric

    raise fdrtd.server.exceptions.NotAvailable(accumulator_name)



import binascii as _binascii
import secrets as _secrets
import hashlib as _hashlib

from fdrtd.plugins.simon.caches.cache import Cache
from fdrtd.plugins.simon.caches.additive import CacheAdditive
from fdrtd.plugins.simon.caches.functional import CacheFunctional
from fdrtd.plugins.simon.microprotocols.microprotocol import Microprotocol
from fdrtd.plugins.simon.crypto.x25519x448 import X25519


class MicroprotocolSetIntersection(Microprotocol):

    def __init__(self, microservice, properties, myself):
        super().__init__(microservice, properties, myself)

        self.secret_key = _secrets.token_bytes(32)

        self.n = len(properties['nodes'])

        self.register_cache('input', Cache())
        self.register_stage(0, ['input'], self.stage_0)

        for i in range(self.n-1):
            self.register_cache('stage{}'.format(i+1), Cache())
            self.register_stage(i+1, ['stage{}'.format(i+1)], self.stage_i)

        self.register_cache('stage{}'.format(self.n), CacheFunctional(lambda x, y: set(x).intersection(set(y)), self.n, self.n))
        self.register_stage(self.n, ['stage{}'.format(self.n)], self.stage_n)

        self.register_cache('reverse', Cache())
        self.register_stage(self.n+1, ['reverse'], self.stage_reverse)

        self.register_cache('samples', CacheAdditive(minimum=self.n))

        self.register_cache('final', Cache())
        self.register_stage(self.n+2, ['final', 'samples'], self.stage_final)

        self.sizes = [0] * self.n
        self.lut = {}
        self.input_lut = {}

    def stage_0(self, args):
        self.network.broadcast(args['input']['samples'], 'samples')
        encrypted = []
        for item in args['input']['set']:
            enc = _hashlib.sha3_256(item.encode("utf-8")).hexdigest()
            self.input_lut[enc] = item
            encrypted.append(enc)
        return self.stage_i({'stage': 0, 'stage0': encrypted})

    def stage_i(self, args):
        stage = args['stage']
        xh = args['stage{}'.format(stage)]
        self.sizes[(self.network.myself-stage) % self.n] = len(xh)
        xxh = self.encrypt(xh)
        if stage == self.network.myself:
            for i in range(len(xxh)):
                self.lut[xxh[i]] = xh[i]
        if stage+1 < self.n:
            self.network.send_to_next_node(xxh, 'stage{}'.format(stage+1))
        else:
            self.network.broadcast(xxh, 'stage{}'.format(stage+1))
        return stage+1, None

    def stage_n(self, args):
        stage = args['stage']
        xxh = args['stage{}'.format(stage)]
        if self.network.myself == self.n-1:
            self.stage_reverse({'stage': stage, 'reverse': xxh})
            return stage + 2, None
        return stage + 1, None

    def stage_reverse(self, args):
        xh = [self.lut[encrypted] for encrypted in args['reverse']]
        if self.network.myself == 0:
            self.network.broadcast(xh, 'final')
        else:
            self.network.send_to_previous_node(xh, 'reverse')
        return args['stage'] + 1, None

    def stage_final(self, args):
        return -1, {'inputs': self.n,
                    'result': {
                        'samples': args['samples'],
                        'size_data': self.sizes,
                        'size_intersection': len(args['final']),
                        'intersection': [self.input_lut[enc] for enc in args['final']]}}

    def encrypt(self, data):
        x25519 = X25519(self.secret_key)
        return [_binascii.hexlify(x25519.encrypt(_binascii.unhexlify(d))).decode('utf-8') for d in data]



from fdrtd.plugins.simon.caches.cache import Cache
from fdrtd.plugins.simon.microprotocols.microprotocol import Microprotocol
from fdrtd.plugins.simon.accumulators.accumulator_statistics_univariate import AccumulatorStatisticsUnivariate


class MicroprotocolStatisticsUnivariate(Microprotocol):

    def __init__(self, microservice, properties, myself):
        super().__init__(microservice, properties, myself)

        self.n = self.network.count

        self.digits_before = properties['parameters'].get('digits_before', 12)
        self.digits_after = properties['parameters'].get('digits_after', 12)

        self.register_cache('input', Cache())
        self.register_cache('uncentered_0', Cache())
        self.register_cache('uncentered_1', Cache())
        self.register_cache('uncentered_2', Cache())
        self.register_cache('uncentered_3', Cache())
        self.register_cache('uncentered_4', Cache())
        self.register_cache('uncentered_5', Cache())
        self.register_cache('samples', Cache())
        self.register_cache('geometric', Cache())
        self.register_cache('harmonic', Cache())
        self.register_cache('minmax', Cache())

        self.register_stage(0, ['input'], self.stage_0)
        self.register_stage(1, ['uncentered_0', 'uncentered_1', 'uncentered_2', 'uncentered_3', 'uncentered_4',
                                'uncentered_5', 'samples', 'geometric', 'harmonic', 'minmax'], self.stage_1)

        self.input = None
        self.result_cache = AccumulatorStatisticsUnivariate()

    def stage_0(self, args):
        self.input = AccumulatorStatisticsUnivariate.deserialize(args['input'])
        self.start_pipeline('SecureSum', 'uncentered_0', self.input.moments.uncentered[0])
        self.start_pipeline('SecureSum', 'uncentered_1', self.input.moments.uncentered[1])
        self.start_pipeline('SecureSum', 'uncentered_2', self.input.moments.uncentered[2])
        self.start_pipeline('SecureSum', 'uncentered_3', self.input.moments.uncentered[3])
        self.start_pipeline('SecureSum', 'uncentered_4', self.input.moments.uncentered[4])
        self.start_pipeline('SecureSum', 'uncentered_5', self.input.moments.uncentered[5])
        self.start_pipeline('SecureSum', 'samples', self.input.samples)
        self.start_pipeline('SecureSum', 'geometric', self.input.geometric)
        self.start_pipeline('SecureSum', 'harmonic', self.input.harmonic)
        self.start_pipeline('MinimumMaximum', 'minmax', [{'minimum': self.input.minimum, 'maximum': self.input.maximum}])
        return 1, None

    def stage_1(self, args):
        self.result_cache.moments.uncentered[0] = args['uncentered_0']['sum']
        self.result_cache.moments.uncentered[1] = args['uncentered_1']['sum']
        self.result_cache.moments.uncentered[2] = args['uncentered_2']['sum']
        self.result_cache.moments.uncentered[3] = args['uncentered_3']['sum']
        self.result_cache.moments.uncentered[4] = args['uncentered_4']['sum']
        self.result_cache.moments.uncentered[5] = args['uncentered_5']['sum']
        self.result_cache.samples = args['samples']['sum']
        self.result_cache.geometric = args['geometric']['sum']
        self.result_cache.harmonic = args['harmonic']['sum']
        self.result_cache.minimum = args['minmax']['minimum']
        self.result_cache.maximum = args['minmax']['maximum']
        self.result_cache.moments.samples = self.result_cache.samples
        self.result_cache.moments.center = 0.0
        self.result_cache.moments.centered = [x for x in self.result_cache.moments.uncentered]
        self.result_cache.moments.set_center(self.result_cache.moments.uncentered[0] / self.result_cache.samples)
        return -1, {'inputs': self.n, 'result': self.result_cache.evaluate_to_dict()}



class Accumulator:

    def evaluate_to_dict(self):
        functions = [func for func in dir(self) if callable(getattr(self, func)) and 'get_' in func]
        return {function[4:]: getattr(self, function)() for function in functions}



from fdrtd.plugins.simon.caches.cache import Cache
from fdrtd.plugins.simon.caches.functional import CacheFunctional
from fdrtd.plugins.simon.caches.additive import CacheAdditive
from fdrtd.plugins.simon.microprotocols.microprotocol import Microprotocol


class MicroprotocolMinimumMaximum(Microprotocol):

    def __init__(self, microservice, properties, myself):
        super().__init__(microservice, properties, myself)

        self.n = self.network.count

        self.intermediate = 0

        self.register_cache('input', Cache())
        self.register_cache('samples', CacheAdditive(minimum=self.n))
        self.register_cache('minimum', CacheFunctional(lambda x, y: x if x < y else y, minimum=self.n))
        self.register_cache('maximum', CacheFunctional(lambda x, y: x if x > y else y, minimum=self.n))

        self.register_stage(0, ['input'], self.stage_0)
        self.register_stage(1, ['samples', 'minimum', 'maximum'], self.stage_1)

    def stage_0(self, args):
        self.network.broadcast(args['input']['samples'], 'samples')
        self.network.broadcast(args['input']['minimum'], 'minimum')
        self.network.broadcast(args['input']['maximum'], 'maximum')
        return 1, None

    def stage_1(self, args):
        return -1, {'inputs': self.n,
                    'result': {
                        'samples': args['samples'],
                        'minimum': args['minimum'],
                        'maximum': args['maximum']}}



from fdrtd.plugins.simon.microprotocols.create import create as create_microprotocol
from fdrtd.plugins.simon.accumulators.create import create as create_accumulator


class TaskSimon:

    def __init__(self, microservice, network, microprotocol, parameters, task_id, parent):

        self.microservice = microservice
        self.task_id = task_id
        self.nodes = network['nodes']
        self.myself = network['myself']
        self.microprotocol_name = microprotocol
        self.microprotocol_class = create_microprotocol(microprotocol)
        self.microprotocol = None
        self.parent = parent
        self.parameters = parameters
        self.accumulator = create_accumulator(microprotocol)()

    def invite(self):
        ret = {'task_id': self.task_id,
               'microprotocol': self.microprotocol_name,
               'invitation_id': None,
               'parameters': self.parameters,
               'parent': self.parent}
        return ret

    def input(self, data):
        if isinstance(data, list):
            for datum in data:
                self.accumulator.update(datum)
        else:
            self.accumulator.update(data)
        return None

    def start(self):
        self.accumulator.finalize()
        args = {'task': {
            'microprotocol': self.microprotocol_name,
            'parent': self.parent,
            'parameters': self.parameters,
            'nodes': [node for node in self.nodes]},
            'message': {
                'token': 'input',
                'sender': self.myself,
                'receiver': self.myself,
                'body': self.accumulator.serialize()}}
        self.peer_to_peer(args)
        return None

    def progress(self):
        return None

    def result(self):
        return self.microprotocol.get_result()

    def peer_to_peer(self, body=None):
        properties = {**body['task'], 'task_id': self.task_id}
        message = body['message']

        if self.microprotocol is None:
            self.microprotocol = self.microprotocol_class(self.microservice, properties, message['receiver'])

        self.parent = body['task']['parent']
        self.parameters = body['task']['parameters']

        self.microprotocol.process(message)
        return None



from fdrtd.plugins.simon.microservice import MicroserviceSimon


def fdrtd_register(registry):

    registry.register(
        {
            "namespace": "fdrtd",
            "protocol": "Simon",
            "version": "0.5.3"
        },
        MicroserviceSimon()
    )



import secrets as _secrets

from fdrtd.plugins.simon.caches.cache import Cache
from fdrtd.plugins.simon.caches.additive import CacheAdditive
from fdrtd.plugins.simon.caches.functional import CacheFunctional
from fdrtd.plugins.simon.microprotocols.microprotocol import Microprotocol
from fdrtd.plugins.simon.accumulators.accumulator_statistics_contingency import AccumulatorStatisticsContingency


class MicroprotocolStatisticsContingencyVertical(Microprotocol):

    def __init__(self, microservice, properties, myself):
        super().__init__(microservice, properties, myself)

        self.secret_key = _secrets.token_bytes(32)

        self.register_cache('input', Cache())
        self.register_cache('keysx', Cache())
        self.register_cache('keysy', Cache())
        self.register_cache('checkpoint2', CacheAdditive(minimum=2))
        self.register_cache('checkpoint3', CacheAdditive(minimum=2))

        self.register_stage(0, ['input'], self.stage_0)
        self.register_stage(1, ['keysx', 'keysy'], self.stage_1)
        self.register_stage(2, ['checkpoint2'], self.stage_2)
        self.register_stage(3, ['checkpoint3'], self.stage_3)

        self.input = None
        self.result_cache = AccumulatorStatisticsContingency()

        self.dictionary = {}
        self.inverted = {}
        self.keysx = []
        self.keysy = []
        self.keys = {}

    def stage_0(self, args):

        self.dictionary = args['input']['dictionary']
        self.inverted = {}
        for key in self.dictionary:
            value = self.dictionary[key]
            if value not in self.inverted:
                self.inverted[value] = []
            self.inverted[value].append(key)

        if self.network.myself == 0:
            self.network.broadcast(self.inverted.keys(), 'keysx')

        if self.network.myself == 1:
            self.network.broadcast(self.inverted.keys(), 'keysy')

        return 1, None

    def stage_1(self, args):

        self.keysx = args['keysx']
        self.keysy = args['keysy']

        for keyx in self.keysx:
            for keyy in self.keysy:
                tag = 'psisize_{}_{}'.format(keyx, keyy)
                self.keys[tag] = (keyx, keyy)
                self.register_cache(tag, Cache())

        self.network.broadcast(0, 'checkpoint2')
        return 2, None

    def stage_2(self, args):
        self.register_stage(4, [*self.keys.keys()], self.stage_final)
        self.network.broadcast(0, 'checkpoint3')
        return 3, None

    def stage_3(self, args):
        for key in self.keys:
            keyx, keyy = self.keys[key]
            if self.network.myself == 0:
                private_set = self.inverted[keyx]
            else:
                private_set = self.inverted[keyy]
            self.start_pipeline('SetIntersectionSize', key, [private_set])
        return 4, None

    def stage_final(self, args):

        table = {}
        mode = None
        max = -1
        for tag in args:
            if tag == 'stage':
                continue
            if tag == 'samples':
                continue
            keyx, keyy = self.keys[tag]
            if keyx not in table:
                table[keyx] = {}
            table[keyx][keyy] = args[tag]['size_intersection']
            if args[tag]['size_intersection'] > max:
                max = args[tag]['size_intersection']
                mode = self.keys[tag]

        return -1, {'inputs': 2, 'result': {'mode': mode, 'table': table}}



from fdrtd.plugins.simon.accumulators.accumulator import Accumulator


class AccumulatorStatisticsMoments(Accumulator):

    def __init__(self, _=None):
        self.samples = 0
        self.center = 0.0
        self.uncentered = [0.0, 0.0, 0.0, 0.0, 0.0, 0.0]
        self.centered = [0.0, 0.0, 0.0, 0.0, 0.0, 0.0]

    def serialize(self):
        return {'samples': self.samples,
                'center': self.center,
                'uncentered': self.uncentered,
                'centered': self.centered}

    @staticmethod
    def deserialize(dictionary):
        accumulator = AccumulatorStatisticsMoments()
        accumulator.samples = dictionary['samples']
        accumulator.center = dictionary['center']
        accumulator.uncentered = dictionary['uncentered']
        accumulator.centered = dictionary['centered']
        return accumulator

    def add(self, other):

        if self.samples == 0:
            self.center = other.center

        self.samples += other.samples

        for k in range(6):
            self.uncentered[k] += other.uncentered[k]
            self.centered[k] += other.recenter(k, self.center)

    def update(self, datum):

        if self.samples == 0:
            self.center = datum

        self.samples = self.samples + 1

        temp = datum
        temp2 = temp * temp
        temp3 = temp2 * temp

        self.uncentered[0] += temp
        self.uncentered[1] += temp2
        self.uncentered[2] += temp3
        self.uncentered[3] += temp2 * temp2
        self.uncentered[4] += temp2 * temp3
        self.uncentered[5] += temp3 * temp3

        temp = datum - self.center
        temp2 = temp * temp
        temp3 = temp2 * temp

        self.centered[0] += temp
        self.centered[1] += temp2
        self.centered[2] += temp3
        self.centered[3] += temp2 * temp2
        self.centered[4] += temp2 * temp3
        self.centered[5] += temp3 * temp3

    def encrypt_data_for_upload(self, nonce, power=1):
        uncentered = [nonce.encrypt_numerical(self.uncentered[i], power=(i+1)*power) for i in range(6)]
        centered = [nonce.encrypt_numerical(self.centered[i], power=(i+1)*power) for i in range(6)]
        return {'samples': self.samples,
                'center': nonce.encrypt_numerical(self.center, power=power),
                'uncentered': uncentered,
                'centered': centered}

    def set_center(self, center):
        recentered = [self.recenter(k, center) for k in range(6)]
        self.centered = recentered
        self.center = center

    def recenter(self, k, center):
        pascal = [[1.0], [1.0, 1.0], [1.0, 2.0, 1.0], [1.0, 3.0, 3.0, 1.0], [1.0, 4.0, 6.0, 4.0, 1.0],
                  [1.0, 5.0, 10.0, 10.0, 5.0, 1.0], [1.0, 6.0, 15.0, 20.0, 15.0, 6.0, 1.0]]

        recentered = pow(self.center - center, k + 1) * self.samples
        for j in range(k + 1):
            recentered += self.centered[k - j] * (pascal[k + 1][j] * pow(self.center - center, j))

        return recentered

    def get_samples(self):
        return self.samples

    def get_center(self):
        return self.center

    def get_uncentered(self, k):
        return self.uncentered[k]

    def get_centered(self, k):
        return self.centered[k]



import math as _math

from fdrtd.plugins.simon.accumulators.accumulator import Accumulator
from fdrtd.plugins.simon.accumulators.accumulator_statistics_univariate import AccumulatorStatisticsUnivariate


class AccumulatorStatisticsBivariate(Accumulator):

    def __init__(self, _=None):
        self.accumulator_x = AccumulatorStatisticsUnivariate()
        self.accumulator_y = AccumulatorStatisticsUnivariate()
        self.accumulator_xy = AccumulatorStatisticsUnivariate()

    def serialize(self):
        return {'accumulator_x': self.accumulator_x.serialize(),
                'accumulator_y': self.accumulator_y.serialize(),
                'accumulator_xy': self.accumulator_xy.serialize()}

    @staticmethod
    def deserialize(dictionary):
        accumulator = AccumulatorStatisticsBivariate()
        accumulator.accumulator_x = AccumulatorStatisticsUnivariate.deserialize(dictionary['accumulator_x'])
        accumulator.accumulator_y = AccumulatorStatisticsUnivariate.deserialize(dictionary['accumulator_y'])
        accumulator.accumulator_xy = AccumulatorStatisticsUnivariate.deserialize(dictionary['accumulator_xy'])
        return accumulator

    def add(self, other):
        self.accumulator_x.add(other.accumulator_x)
        self.accumulator_y.add(other.accumulator_y)
        self.accumulator_xy.add(other.accumulator_xy)

    def update(self, data):
        (x, y) = data
        self.accumulator_x.update(x)
        self.accumulator_y.update(y)
        self.accumulator_xy.update(x*y)

    def finalize(self):
        self.accumulator_x.finalize()
        self.accumulator_y.finalize()
        self.accumulator_xy.finalize()

    def encrypt_data_for_upload(self, nonce):
        return {'accumulator_x': self.accumulator_x.encrypt_data_for_upload(nonce),
                'accumulator_y': self.accumulator_y.encrypt_data_for_upload(nonce),
                'accumulator_xy': self.accumulator_xy.encrypt_data_for_upload(nonce, power=2)}

    @staticmethod
    def decrypt_result_from_download(encrypted, nonce):
        decryption_powers = {'samples': 0, 'covariance_mle': 2, 'covariance': 2, 'correlation_coefficient': 0,
                             'regression_slope': 0, 'regression_interceipt': 1, 'regression_slope_only': 0}
        return nonce.decrypt_dictionary_numerical(encrypted, decryption_powers)

    def get_samples(self):
        return self.accumulator_xy.get_samples()

    def get_covariance_mle(self):
        return self.accumulator_xy.get_mean() - self.accumulator_x.get_mean() * self.accumulator_y.get_mean()

    def get_covariance(self):
        return self.get_covariance_mle() / (1.0 - 1.0 / self.accumulator_xy.get_samples())

    def get_correlation_coefficient(self):
        return self.get_covariance() / _math.sqrt(self.accumulator_x.get_variance() * self.accumulator_y.get_variance())

    def get_regression_slope(self):
        return self.get_covariance() / self.accumulator_x.get_variance()

    def get_regression_interceipt(self):
        return self.accumulator_y.get_mean() - self.get_regression_slope() * self.accumulator_x.get_mean()

    def get_regression_slope_only(self):
        return self.accumulator_xy.get_mean() / self.accumulator_x.calculate_raw_moment(2)



from fdrtd.plugins.simon.accumulators.accumulator import Accumulator


class AccumulatorBasicMinimumMaximum(Accumulator):

    def __init__(self, _=None):
        self.samples = 0
        self.minimum = None
        self.maximum = None

    def serialize(self):
        return {'samples': self.samples,
                'minimum': self.minimum,
                'maximum': self.maximum}

    @staticmethod
    def deserialize(dictionary):
        accumulator = AccumulatorBasicMinimumMaximum
        accumulator.samples = dictionary['samples']
        accumulator.minimum = dictionary['minimum']
        accumulator.maximum = dictionary['maximum']
        return accumulator

    def add(self, other):
        self.samples = self.samples + other.samples
        self.minimum = min(self.minimum, other.minimum)
        self.maximum = max(self.maximum, other.maximum)

    def update(self, data):

        if self.samples == 0:
            self.samples = 1
            self.minimum = data['minimum']
            self.maximum = data['maximum']
            return

        self.samples = self.samples + 1
        self.minimum = min(self.minimum, data['minimum'])
        self.maximum = max(self.maximum, data['maximum'])

    def finalize(self):
        pass

    def get_samples(self):
        return self.samples

    def get_min(self):
        return self.minimum

    def get_max(self):
        return self.maximum
